from typing import Optional, List, Union, Callable
import numpy as np
from qiskit.circuit import Parameter, ParameterVector
from qiskit.circuit.library import PauliFeatureMap 
from qiskit import QuantumCircuit, QuantumRegister
from quple.components.interaction_graphs import interaction_graph
from qiskit.aqua.components.feature_maps import FeatureMap
import operator
import functools

class FeatureMaps01(FeatureMap):
    """
    Mapping data with the second order expansion followed by entangling gates.
    Refer to https://arxiv.org/pdf/1804.11326.pdf for details.
    """

    CONFIGURATION = {
        'name': 'MultiVariableMap',
        'description': 'Second order expansion for feature map',
        'input_schema': {
            '$schema': 'http://json-schema.org/schema#',
            'id': 'Second_Order_Expansion_schema',
            'type': 'object',
            'properties': {
                'depth': {
                    'type': 'integer',
                    'default': 2,
                    'minimum': 1
                },
                'entangler_map': {
                    'type': ['array', 'null'],
                    'default': None
                },
                'entanglement': {
                    'type': 'string',
                    'default': 'full',
                    'oneOf': [
                        {'enum': ['full', 'linear']}
                    ]
                }
            },
            'additionalProperties': False
        }
    }

    def __init__(self, num_qubits, depth=2, degree=1):
        """Constructor.
        Args:
            num_qubits (int): number of qubits
            depth (int): the number of repeated circuits
            entangler_map (list[list]): describe the connectivity of qubits, each list describes
                                        [source, target], or None for full entanglement.
                                        Note that the order is the list is the order of
                                        applying the two-qubit gate.
            entanglement (str): ['full', 'linear'], generate the qubit connectivitiy by predefined
                                topology
            data_map_func (Callable): a mapping function for data x
        """
        #self.validate(locals())
        super().__init__()
        self._num_qubits = num_qubits
        self._feature_dimension = num_qubits
        self._depth = depth
        self._support_parameterized_circuit = False
        self._support_parameterized_circuit = False
        self._degree = degree

    def _build_circuit_template(self):
        return True

    def construct_circuit(self, x, qr=None, inverse=False):
        #logger.info("8 qubits 26 variable feature map is used")
        """
        Construct the second order expansion based on given data.
        Args:
            x (numpy.ndarray): 1-D to-be-transformed data.
            qr (QauntumRegister): the QuantumRegister object for the circuit, if None,
                                  generate new registers with name q.
            inverse (bool): whether or not inverse the circuit
        Returns:
            QuantumCircuit: a quantum circuit transform data x.
        """

        if qr is None:
            qr = QuantumRegister(self._num_qubits, name='q')

        qc = QuantumCircuit(qr)

        entangler_map = []
        for i in range(0, self._num_qubits - 1, 2):
            entangler_map.append([i, i + 1])
        for i in range(0, self._num_qubits - 2, 2):
            entangler_map.append([i + 1, i + 2])

        for d in range(self._depth):
            for i in range(self._num_qubits):
                qc.h(qr[i])
                qc.rz(x[i], qr[i])
                if self._degree>0:
                   qc.ry(x[i]**self._degree,qr[i])
            for src, targ in entangler_map:
                qc.cx(qr[src], qr[targ])
                qc.rz(((x[src] + x[targ])**self._degree)/np.pi, qr[targ])
                qc.cx(qr[src], qr[targ])
        if inverse:
            qc = qc.inverse()
        return qc

class FeatureMapDev(PauliFeatureMap):
    def pauli_evolution(self, pauli_string, time):
        """Get the evolution block for the given pauli string."""
        # for some reason this is in reversed order
        pauli_string = pauli_string[::-1]

        # trim the pauli string if identities are included
        trimmed = []
        indices = []
        for i, pauli in enumerate(pauli_string):
            if pauli != 'I':
                trimmed += [pauli]
                indices += [i]

        evo = QuantumCircuit(len(pauli_string))

        if len(trimmed) == 0:
            return evo

        def basis_change(circuit, inverse=False):
            # do not change basis if only first order pauli operator
            if len(pauli_string) == 1:
                return
            for i, pauli in enumerate(pauli_string):
                if pauli == 'X':
                    circuit.h(i)
                elif pauli == 'Y':
                    circuit.rx(-np.pi / 2 if inverse else np.pi / 2, i)

        def cx_chain(circuit, inverse=False):
            num_cx = len(indices) - 1
            for i in reversed(range(num_cx)) if inverse else range(num_cx):
                circuit.cx(indices[i], indices[i + 1])

        basis_change(evo)
        cx_chain(evo)
        if len(pauli_string) == 1:
            if pauli_string[0] == 'Z':
                evo.rz(time, indices[-1])
            elif pauli_string[0] == 'X':
                evo.rx(time, indices[-1])
            elif pauli_string[0] == 'Y':
                evo.ry(time, indices[-1])
        else:
            evo.rz(time, indices[-1])
        cx_chain(evo, inverse=True)
        basis_change(evo, inverse=True)
        return evo
    
    def get_entangler_map(self, rep_num: int, 
                          block_num: int, num_block_qubits: int) :
        strategy = self.entanglement
        if isinstance(strategy, str):
            return interaction_graph[strategy](self.num_qubits, num_block_qubits)
        elif callable(strategy):
            return strategy(self.num_qubits, num_block_qubits)
        elif isinstance(strategy, list):
            if all(isinstance(strat, str) for strat in strategy):
                return interaction_graph[strategy[block_num]](self.num_qubits, num_block_qubits)
            elif all(callable(strat) for strat in strategy):
                return strategy[block_num](self.num_qubits, num_block_qubits)
        else:
            raise ValueError('invalid entangle strategy: {}'.format(strategy))


class FeatureMaps02(PauliFeatureMap):

    def __init__(self,
                 feature_dimension: Optional[int] = None,
                 reps: int = 2,
                 entanglement: Union[str, List[List[int]], Callable[[int], List[int]]] = 'full',
                 paulis: Optional[List[str]] = None,
                 data_map_func: Optional[Callable[[np.ndarray], float]] = None,
                 parameter_prefix: str = 'x',
                 insert_barriers: bool = False,
                 degree: int= 2
                 ) -> None:

        super().__init__(feature_dimension=feature_dimension,
                         reps=reps,
                         entanglement=entanglement,
                         paulis=paulis,
                         data_map_func=data_map_func,
                         parameter_prefix=parameter_prefix,
                         insert_barriers=insert_barriers)

        self._degree = degree


    def _custom_data_map_func(self, x):
        if len(x) == 1:
            coeff = x[0]
        else:
            coeff = functools.reduce(lambda m, n: (m + n)._apply_operation(operator.pow, self._degree)/(np.pi), x)
        return coeff

    def pauli_block(self, pauli_string):
        """Get the Pauli block for the feature map circuit."""
        params = ParameterVector('_', length=len(pauli_string))
        
        time = self._custom_data_map_func(np.asarray(params))
        return self.pauli_evolution(pauli_string, time)

    def pauli_evolution(self, pauli_string, time):
        """Get the evolution block for the given pauli string."""
        # for some reason this is in reversed order
        pauli_string = pauli_string[::-1]

        # trim the pauli string if identities are included
        trimmed = []
        indices = []
        for i, pauli in enumerate(pauli_string):
            if pauli != 'I':
                trimmed += [pauli]
                indices += [i]

        evo = QuantumCircuit(len(pauli_string))

        if len(trimmed) == 0:
            return evo

        def basis_change(circuit, inverse=False):
            for i, pauli in enumerate(pauli_string):
                if pauli == 'X':
                    circuit.h(i)
                elif pauli == 'Y':
                    if self._degree>0:
                        circuit.ry(time._apply_operation(operator.pow, self._degree), i)

        def cx_chain(circuit, inverse=False):
            num_cx = len(indices) - 1
            for i in reversed(range(num_cx)) if inverse else range(num_cx):
                circuit.cx(indices[i], indices[i + 1])

        basis_change(evo)
        cx_chain(evo)
        evo.p(time, indices[-1])
        cx_chain(evo, inverse=True)

        
        return evo            


import os
import time
import datetime
import logging

import numpy as np
import tensorflow as tf

import quple

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class DefaultDict(dict):
    def __missing__(self, key):
        return key
    
class VQCLogger(tf.keras.callbacks.Callback):
    """Logger for the variational quantum classifier (VQC)
    
    """
    def __init__(self, log_dir='./logs', 
                 filename='{encoding_circuit}_{encoding_map}_'
                 '{variational_circuit}_{n_qubit}_qubit_'
                 '{activation}_{optimizer}_train_{train_size}_'
                 'val_{val_size}_test_{test_size}_epoch_{epochs}_'
                 'batch_size_{batch_size}_{time}',
                 keys=None,
                 stream_level=logging.INFO,
                 file_level=logging.DEBUG,
                 formatter=None,
                 save_npz=True,
                 save_weights=True,
                 roc_plot=True):
        """Creates the VQC logger
        Args:
            log_dir: str
                Name of logging directory.
            filename: str
                File name prefix for logged items.
            keys: str
                Keys of epoch results to save.
            stream_level: str
                Stream level for the logger.
            file_level: str
                File level for the logging file.
            formatter: str
                Formatter for logging.
            save_npz: boolean; default=True
                Whether to save VQC specs, epoch results and test results
                to a npz file.
            save_weights boolean; default=True
                Whether to save the weights of the model.
            roc_plot: boolean; default=True
                Whether to save a roc plot for the test result.
         
        """
        self.log_dir = log_dir
        self.filename = filename
        self.keys = keys
        self.stream_level = stream_level
        self.file_level = file_level
        self.stream_handler = None
        self.file_handler = None
        self.formatter = formatter or logging.Formatter('%(asctime)s [%(threadName)-12.12s]'
                                                         '[%(levelname)-5.5s]  %(message)s')
        self.time = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d_%H-%M-%S')
        self.logger_is_set = False
        self.save_npz = save_npz
        self.save_weights = save_weights
        self.roc_plot = roc_plot
        super(VQCLogger, self).__init__() 
        
    def setup_logger(self):
        if self.logger_is_set:
            return
        # setup stream handler
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(self.stream_level)
        stream_handler.setFormatter(self.formatter)
        
        # setup file handler
        attrib = self.model.attributes
        attrib['time'] = self.time
        attrib = DefaultDict(attrib)
        self.formatted_filename = os.path.join(self.log_dir, self.filename.format(**attrib))
        log_filename = self.formatted_filename + '.log'
        os.makedirs(os.path.dirname(log_filename), exist_ok=True)
        file_handler = logging.FileHandler(log_filename)
        file_handler.setLevel(self.file_level)
        file_handler.setFormatter(self.formatter)
        self.stream_handler = stream_handler
        self.file_handler = file_handler
        
        #setup logger
        logger.addHandler(self.stream_handler)
        logger.addHandler(self.file_handler)
        self.logger_is_set = True
        self.print_attributes()
        
    def print_attributes(self):
        attrib = self.model.attributes
        logger.info('######## Executing VQC with the following attributes ########')
        logger.info('Feature Dimension: {}'.format(attrib['feature_dimension']))            
        logger.info('Number of Qubits: {}'.format(attrib['n_qubit']))
        logger.info('Qubit Layout: {}'.format(list(attrib['qubits'])))
        logger.info('Encoding Circuit: {}'.format(attrib['encoding_circuit']))
        logger.info('Encoding Map: {}'.format(attrib['encoding_map']))         
        logger.info('Variational Circuit: {}'.format(attrib['variational_circuit']))            
        logger.info('Circuit Parameters: {}'.format(list(attrib['circuit_parameters'])))
        logger.info('Number of Parameters: {}'.format(attrib['num_parameters']))            
        logger.info('Optimizer: {}'.format(attrib['optimizer']))
        logger.info('Metrics: {}'.format(list(attrib['metrics']))) 
        logger.info('Loss Function: {}'.format(list(attrib['loss']))) 
        logger.info('Activation Function: {}'.format(attrib['activation'])) 
        logger.info('Training Size: {}'.format(attrib['train_size']))
        logger.info('Validation Size: {}'.format(attrib['val_size']))
        logger.info('Test Size: {}'.format(attrib['test_size'])) 
        logger.info('Batch Size: {}'.format(attrib['batch_size'])) 
        logger.info('Number of Epochs: {}'.format(attrib['epochs'])) 
        
        logger.debug('Circuit Diagram for Encoding Circuit:\n{}'.format(self.model.encoding_circuit))
        logger.debug('Circuit Diagram for Variational Circuit:\n{}'.format(self.model.variational_circuit))
                    
    def reset_logger(self):
        self.file_handler.close()
        logger.removeHandler(self.stream_handler)
        logger.removeHandler(self.file_handler)
        self.logger_is_set = False
    
    def on_train_begin(self, logs=None):
        self.setup_logger()  
        self.train_result = []
        logger.info('######## Training Begins ########')
        logger.info('Number of samples for Training: {}'.format(self.params['samples']))
        logger.info('Number of Epochs: {}'.format(self.params['epochs']))
        logger.info('Batch Size: {}'.format(self.params['batch_size']))
        
    def on_train_end(self, logs=None):
        logger.info('######## Training Ends ########')
        weights = self.model.get_weights()[0]
        logger.info('Model weights: \n{}'.format(weights))
        if self.save_npz:
            filename = self.formatted_filename +'.npz'
            self.savez(filename, 
                     train_result=self.train_result,
                     model_weights=weights,
                     **self.model.attributes)
        
        if self.save_weights:
            weights_filename = self.formatted_filename + '_model_weights'
            self.model.save_weights(weights_filename)

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}  
        
        # log all keys in logs if not specified
        if self.keys is None:
              self.keys = sorted(logs.keys())
        
        # set value to np.NaN if value is not available in last epoch
        if self.model.stop_training:
            logs = dict((k, logs[k]) if k in logs else (k, np.NaN) for k in self.keys)       
                    
        logger.debug('######## Epoch {} ########'.format(epoch))
        for key in logs:
            logger.debug('{}: {}'.format(key, logs[key]))
        logs['epoch'] = epoch    
        self.train_result.append(logs)
        
    def on_test_begin(self, logs=None):
        # only log when it's calling model.evaluate
        if self.params.get('do_validation', None):
            return
        logger.info('######## Test Begins ########')
        logger.info('Number of samples for Testing: {}'.format(self.params['samples']))
        logger.info('Number of Epochs: {}'.format(self.params['epochs']))
        logger.info('Batch Size: {}'.format(self.params['batch_size']))
    
    def savez(self, filename, **kwargs):
        temp = {}
        if os.path.exists(filename):
            temp = dict(np.load(filename, allow_pickle=True))
        temp.update(kwargs)
        np.savez(filename, **temp)
        
    def on_test_end(self, logs=None):
        logs = logs or {} 
        # only log when it's calling model.evaluate
        if self.params.get('do_validation', None):
            return
        for key in logs:
            logger.info('{}: {}'.format(key, logs[key]))
        logger.info('*Note: for tensorflow version < 2.3.0, '
                   'the test results are not loaded into logs')
        logger.info('######## Test Ends ########')
        
        if self.save_npz and logs:
            filename = self.formatted_filename +'.npz'
            self.savez(filename, test_result=logs)
 
                    

    def log_roc_curve(self, fpr, tpr, auc):
        logger.debug('######## ROC curve information ########')
        logger.debug('fpr:\n{}'.format(fpr))
        logger.debug('tpr:\n{}'.format(tpr))
        logger.debug('auc:\n{}'.format(auc))
        roc_result = {'fpr':fpr, 'tpr':tpr, 'auc':auc}
        if self.save_npz:
            filename = self.formatted_filename +'.npz'
            self.savez(filename, roc_result=roc_result)
        if self.roc_plot:
            import matplotlib
            matplotlib.use('Agg') 
            plt = quple.utils.utils.plot_roc_curve(fpr, tpr)
            filename = self.formatted_filename+'_roc_curve'
            plt.tight_layout()
            plt.savefig(filename+'.png')
            plt.savefig(filename+'.eps')
                    

    
    


from quple import QuantumCircuit

def construct_bell_circuit():
    """
    Quantum circuit to create the Bell state:
    |\Phi ^{+}\rangle ={\frac  {1}{{\sqrt  {2}}}}(|0\rangle _{A}\otimes |0\rangle _{B}+|1\rangle _{A}\otimes |1\rangle _{B})
    """
    cq = QuantumCircuit(2, name='BellCircuit')
    cq.H(0)
    cq.CNOT((0,1))
    return cq

bell_circuit = construct_bell_circuit()



from abc import ABC, abstractmethod
from typing import Optional, List, Union, Callable, Tuple
import numpy as np

import cirq

from quple import ParameterisedCircuit, TemplateCircuitBlock
from quple.data_encoding.encoding_maps import encoding_map_registry

class EncodingCircuit(ParameterisedCircuit, ABC):
    '''The data encoding circuit (or state preparation circuit)
    
    A quantum circuit for encoding classical data into a quantum state by embedding the
    data in the parameters of unitary operations on the circuit qubits using a suitable encoding function.
    
    If the qubit encoding method is used, the number of qubits in the encoding circuit should match the 
    feature dimesion of the input data.
    
    '''
    def __init__(self, feature_dimension:int, copies:int=1,
                rotation_blocks:Optional[Union[str, cirq.Gate, Callable, TemplateCircuitBlock,
                                                               List[str],List[cirq.Gate],List[Callable],
                                                               List[TemplateCircuitBlock]]]='H',
                entanglement_blocks:Optional[Union[str, cirq.Gate, Callable, TemplateCircuitBlock,
                                               List[str],List[cirq.Gate],List[Callable],
                                               List[TemplateCircuitBlock]]]=None,
                entangle_strategy:Optional[Union[str,List[str], Callable[[int,int],List[Tuple[int]]],
                                                 List[Callable[[int,int],List[Tuple[int]]]]]]=None,
                encoding_map: Optional[Union[str, Callable]]=None,
                parameter_symbol:str='x',
                name:str='EncodingCircuit',
                flatten_circuit=True,
                *args, **kwargs):
        '''Create a new Pauli expansion circuit.
        Args:
            feature_dimension: dimension of data to be encoded (=number of qubits in the circuit)
            copies: the number of repetition of the encoding circuit
            encoding_map: data mapping function from R^(feature_dimension) to R
            parameter_symbol: the symbol prefix for the encoded data
            name: name of circuit
            flatten_circuit: whether or not to flatten the circuit
        '''
        self._feature_dimension = feature_dimension
        self.encoding_map = encoding_map
        super().__init__(n_qubit=feature_dimension, copies=copies,
                         rotation_blocks=rotation_blocks,
                         entanglement_blocks=entanglement_blocks,
                         entangle_strategy=entangle_strategy,
                         parameter_symbol=parameter_symbol,
                         final_rotation_layer=False, 
                         flatten_circuit=flatten_circuit,
                         name=name, *args, **kwargs)
        
    def get_parameters(self, indices:List[int]):
        '''Obtain parameter symbols corresponding to the array indices
        Args:
            indices: array indices of the parameter
        Return:
            a numpy array of parameter symbols
        Example:    
        >>>  cq.get_parameters([1, 2, 3])
        array([x_1, x_2, x_3], dtype=object)
        '''
        if self.num_param == 0:
            self.new_param(size=self.feature_dimension)
        if isinstance(indices, tuple):
            indices = list(indices)
        return self.parameters[indices]

    def encode_parameters(self, indices:List[int], encoding_map:Callable=None):
        '''Obtain the encoded value for the given parameters using the encoding map
        Args:
            indices: indices of parameter array which corresponds to the indices of qubits in the circuit
        Returns:
            A symbolic expression of the encoded data
        Example:
        >>> from quple.data_encoding import PauliExpansion
        >>> cq = PauliExpansion(feature_dimension=5, copies=1, encoding_map='self_product')
        >>> cq.encode_parameters(indices=[0,2,4])
        >>> 𝑥0𝑥2𝑥4
        '''
        encoding_map = encoding_map or self.encoding_map
        parameters = self.get_parameters(indices)
        encoded_value = encoding_map(parameters)
        return encoded_value
    
    def get_total_parameter_count(self) -> int:
        return self._feature_dimension
    
    @property
    def feature_dimension(self):
        return self._feature_dimension
    
    @property
    def encoding_map(self):
        return self._encoding_map
    
    @encoding_map.setter
    def encoding_map(self, val):
        if val is None:
            val = encoding_map_registry['self_product']
        elif isinstance(val, str):
            val = encoding_map_registry[val]
        self._encoding_map = val


from typing import Callable, Optional
import numpy as np

from quple.data_encoding.general_pauli_encoding import GeneralPauliEncoding

class GeneralPauliZEncoding(GeneralPauliEncoding):
    """The general Pauli Z encoding circuit
    
    An encoding circuit consisting of layers of unitary operators of the
    form exp(iψ(x)Z^{⊗k})H^{⊗n} where ψ is a data encoding function, Z^{⊗k}
    is a k fold tensor product of Pauli Z operator for some k <= n on n qubits,
    and x = (x_1, . . . , x_n ) are the input features to be encoded.
    
    To encode data of feature dimension n, a set of general Pauli operators 
    are chosen to encode the data into an n qubit circuit. Each Pauli operator
    will contribute to a unitary operation \exp(i\sum_{s\in S}ψ_s(x_s)Σ_s)H^{⊗n}
    where s is the indices of a subset of all qubit indices S. For a general Pauli 
    operator of order k, s is a tuple of k elements. 
    
    For example, for k = 1 case S is the set {1, 2, ..., n} and 
    the unitary operation is \exp(i\sum_{j=1}^n ψ_j(x_j)Z_j)H^{⊗n}, 
    where Z_j is the Pauli Z operator acting on the j-th qubit. 
    
    If instead k = 2, then S is a set of 2-tuple of
    qubit indices determined by the interaction graph. For a fully connected
    graph, S is the set of combinations of 2-qubit pairs. Then the unitary operation
    is \exp(i\sum{s=(j, k) in S}ψ_{j,k}(x_j, x_k)Z_j⊗Z_k)H^{⊗n}
    
    The Pauli encoding uses the circuit-centric design with alternating
    rotation and entanglement layer to achieve a strongly entangling circuit. 
    Ref: https://arxiv.org/pdf/1804.00633.pdf    

    Examples:
    >> cq = GeneralPauliZEncoding(feature_dimension=3, z_order=2, copies=1)
    >> cq
    (0, 0): ───H───Rz(pi*x_0)───@──────────────────────@───@──────────────────────@──────────────────────────────
                                │                      │   │                      │
    (0, 1): ───H───Rz(pi*x_1)───X───Rz(pi*<x_0*x_1>)───X───┼──────────────────────┼───@──────────────────────@───
                                                           │                      │   │                      │
    (0, 2): ───H───Rz(pi*x_2)──────────────────────────────X───Rz(pi*<x_0*x_2>)───X───X───Rz(pi*<x_1*x_2>)───X───
    
    Args:
        feature_dimension: int
            Dimension of data to be encoded (=number of qubits in the circuit)
        copies: int
            Number of repetition of the encoding circuit
        encoding_map: function that maps a numpy array to a number
            Data mapping function from R^(feature_dimension) to R
        z_order: int
            Order of pauli z operations to be performed on each circuit block
        name: str
            Name of the encoding circuit
    """
    def __init__(self, feature_dimension: int,
                 copies:int=2, z_order:int=2,
                 encoding_map:Optional[Callable[[np.ndarray], float]] = None,
                 global_shift:bool=False,
                 name:str='GeneralPauliZEncoding', *args, **kwargs):
        '''Creates the general Pauli Z encoding circuit

        Args:
            feature_dimension: int
                Dimension of data to be encoded (=number of qubits in the circuit)
            copies: int
                Number of repetition of the encoding circuit
            z_order: int
                Order of pauli z operations to be performed on each circuit block
            encoding_map: function that maps a numpy array to a number
                Data mapping function from R^(feature_dimension) to R
            global_shift: bool
                Include global shift to RZ gate
            name: str
                Name of the encoding circuit
        '''      
        paulis = []
        for i in range(1, z_order + 1):
            paulis.append('Z' * i)
        super().__init__(feature_dimension, copies=copies, paulis=paulis,
                         encoding_map=encoding_map, global_shift=global_shift,
                         name=name, *args, **kwargs)


import itertools
import numpy as np
from qiskit.circuit import ParameterVector
from qiskit import ClassicalRegister, QuantumCircuit, QuantumRegister
from qiskit.circuit import ParameterVector
from qiskit.aqua.algorithms.classifiers import QSVM 
from qiskit.aqua.components.feature_maps import FeatureMap
import quple
from quple.utils.utils import parallel_run


def construct_circuit(x, feature_map):

    q = QuantumRegister(feature_map.num_qubits, 'q')
    c = ClassicalRegister(feature_map.num_qubits, 'c')
    qc = QuantumCircuit(q, c)

    # write input state from sample distribution
    if isinstance(feature_map, FeatureMap):
        qc += feature_map.construct_circuit(x, q)
    else:
      raise ValueError('Only FeatureMap istance is allowed')
    return qc

def get_qiskit_state_vectors_test(quantum_instance, feature_map, x):
    is_statevector_sim = quantum_instance.is_statevector

    measurement = not is_statevector_sim
    measurement_basis = '0' * feature_map.num_qubits

        
    # build parameterized circuits, it could be slower for building circuit
    # but overall it should be faster since it only transpile one circuit
    feature_map_params = ParameterVector('x', feature_map.feature_dimension)
    parameterized_circuit = QSVM._construct_circuit(
        (feature_map_params, feature_map_params), feature_map, measurement,
        is_statevector_sim=is_statevector_sim)
    parameterized_circuit = quantum_instance.transpile(parameterized_circuit)[0]
    circuits = [parameterized_circuit.assign_parameters({feature_map_params: params})
                for params in x]        
    results = quantum_instance.execute(circuits, had_transpiled=True) 
    statevectors = np.array([results.get_statevector(i) for i in range(len(results.results))])
    return statevectors
    
def assign_parameters(param_values, param_names, feature_map):
  return feature_map.assign_parameters({param_names: param_values})

def get_qiskit_state_vectors(quantum_instance, feature_map, x):
    if not quantum_instance.is_statevector:
        raise ValueError('Quantum instance must be a statevector simulator')
    q = QuantumRegister(feature_map.num_qubits, 'q')
    c = ClassicalRegister(feature_map.num_qubits, 'c')
    qc = QuantumCircuit(q, c)
    
    
    if isinstance(feature_map, QuantumCircuit):
      use_parameterized_circuits = True
    else:
        use_parameterized_circuits = feature_map.support_parameterized_circuit    
        
    if isinstance(feature_map, FeatureMap):
      circuits = parallel_run(construct_circuit, x, itertools.repeat(feature_map))
    else:
      feature_map_params = ParameterVector('x', feature_map.feature_dimension)
      parameterized_circuit = QSVM._construct_circuit(
          (feature_map_params, feature_map_params), feature_map, False,
          is_statevector_sim=True)
      parameterized_circuit = quantum_instance.transpile(parameterized_circuit)[0]
      circuits = parallel_run(assign_parameters, x, itertools.repeat(feature_map_params))    
    results = quantum_instance.execute(circuits, had_transpiled=use_parameterized_circuits) 
    statevectors = np.array([results.get_statevector(i) for i in range(len(results.results))])
    return statevectors
    
    
'''
    def construct_circuit(data):
        return feature_map.construct_circuit(data, q)
    circuits = parallel_run(construct_circuit, x)
    
'''    



from typing import Sequence
import numpy as np

from quple import TemplateCircuitBlock

class RISWAPBlock(TemplateCircuitBlock):
    
    @staticmethod
    def RYY(self, circuit:'quple.ParameterisedCircuit', qubits:Sequence[int], theta):
        circuit.RX(np.pi/2, list(qubits))
        circuit.CX(tuple(qubits))
        circuit.RZ(theta, qubits[1])
        circuit.CX(tuple(qubits))
        circuit.RX(-np.pi/2, list(qubits))
        
    @staticmethod
    def RXX(self, circuit:'quple.ParameterisedCircuit', qubits:Sequence[int], theta):
        circuit.H(list(qubits))
        circuit.CX(tuple(qubits))
        circuit.RZ(theta, qubits[1])
        circuit.CX(tuple(qubits))
        circuit.H(list(qubits))
        

    def build(self, circuit:'quple.ParameterisedCircuit', qubits:Sequence[int]):
        theta = circuit.new_param()
        RISWAPBlock.RXX(theta, qubits)
        RISWAPBlock.RYY(theta, qubits)
    
    @property
    def num_block_qubits(self) -> int:
        return 2
